# Joel Andersson, joel@casadi.org, 2016
from casadi import *
import numpy
def easycall(fcn, *args):
    return fcn(args)[0] if len(args)==1 else fcn(args)
def collocation(dae, tf, nsteps):

    # Degree of interpolating polynomial
    d = 3

    # Get collocation points
    tau_root = collocationPoints(d, 'legendre')

    # Coefficients of the collocation equation
    C = numpy.zeros((d+1,d+1))

    # Coefficients of the continuity equation
    D = numpy.zeros(d+1)

    # Coefficients of the quadrature function
    B = numpy.zeros(d+1)

    # Construct polynomial basis
    for j in range(d+1):
        # Construct Lagrange polynomials to get the polynomial basis at the collocation point
        p = numpy.poly1d([1])
        for r in range(d+1):
            if r != j:
                p *= numpy.poly1d([1, -tau_root[r]]) / (tau_root[j]-tau_root[r])
  
        # Evaluate the polynomial at the final time to get the coefficients of the continuity equation
        D[j] = p(1.0)

        # Evaluate the time derivative of the polynomial at all collocation points to get the coefficients of the continuity equation
        pder = numpy.polyder(p)
        for r in range(d+1):
            C[j,r] = pder(tau_root[r])

        # Evaluate the integral of the polynomial to get the coefficients of the quadrature function
        pint = numpy.polyint(p)
        B[j] = pint(1.0)

    # Continuous time dynamics
    f = Function('f', [dae['x'], dae['z']], [dae['ode'], dae['alg']])

    # Variables for one finite element
    X = [MX.sym('X' + str(k), dae['x'].sparsity()) for k in range(d+1)]
    Z = [MX.sym('Z' + str(k), dae['z'].sparsity()) for k in range(d+1)]

    # Collocation equations and quadratures
    g = []
    h = tf/nsteps
    for j in range(1,d+1):
        # Expression for the state derivative at the collocation point
        xp = 0
        for r in range(d+1):
            xp = xp + C[r,j]*X[r]
      
        # Append collocation equations
        fj, algj = easycall(f, X[j],Z[j])
        g += [h*fj - xp]

        # Append algebraic equations
        g += [algj]
  
    # State at end
    xf = 0
    for j in range(d+1): xf = xf + D[j]*X[j]

    # State at end
    zf = 0
    for j in range(d+1): zf = zf + D[j]*Z[j]
  
    # Rootfinder function, implicitly defines X and Q as a function of X0 and U
    V = []
    for j in range(1,d+1): V += [X[j], Z[j]]

    rfp = Function('rfp', [vertcat(V), X[0], Z[0]], [vertcat(g), xf, zf])
    rf = rootfinder('rf', 'newton', rfp)

    # Discrete time dynamics
    XF = X[0]
    ZF = Z[0]
    for j in range(nsteps):
        _, XF, ZF = easycall(rf, repmat(vertcat((XF,ZF)), d, 1), XF, ZF)
    return Function('F', [X[0], Z[0]], [XF, ZF], ['x0', 'z0'], ['xf', 'zf'])
